#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/tads3/resload.cpp,v 1.2 1999/05/17 02:52:29 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  resload.cpp - resource loader
Function
  
Notes
  
Modified
  10/17/98 MJRoberts  - Creation
*/

#include "t3std.h"
#include "resload.h"

/*
 *   create resource loader 
 */
CResLoader::CResLoader()
{
    /* we have no root directory */
    root_dir_ = 0;
}

/*
 *   create a resource loader given a root directory 
 */
CResLoader::CResLoader(const char *root_dir)
{
    /* remember the root directory for file searches */
    root_dir_ = lib_copy_str(root_dir);
}

/*
 *   delete resource loader 
 */
CResLoader::~CResLoader()
{
    /* free our root directory string */
    lib_free_str(root_dir_);
}

/*
 *   Open a resource file given the resource path 
 */
osfildef *CResLoader::open_res_file(const char *respath)
{
    char filepath[OSFNMAX];
    osfildef *fp;
    
    /* first, look in any embedded resources we have */
    // $$$ to be implemented

    /* 
     *   We didn't find an embedded resource answering to the specified
     *   name, so look for the resource as an external file.  If we have a
     *   root directory, look for the file under that directory;
     *   otherwise, look for it in the current directory.  
     */
    if (root_dir_ != 0)
    {
        char fname[OSFNMAX];
        
        /* get the resource name as a file path */
        os_cvt_url_dir(fname, sizeof(fname), respath, FALSE);

        /* build a full path from the root directory and the resource path */
        os_build_full_path(filepath, sizeof(filepath), root_dir_, fname);
    }
    else
    {
        /* get the resource name as a file path */
        os_cvt_url_dir(filepath, sizeof(filepath), respath, FALSE);
    }

    /* try opening the file */
    fp = osfoprb(filepath, OSFTBIN);

    /* 
     *   if that failed, and we had a root directory to look in, try again
     *   in the current directory 
     */
    if (fp == 0 && root_dir_ != 0)
    {
        /* get the resource name as a simple file path */
        os_cvt_url_dir(filepath, sizeof(filepath), respath, FALSE);

        /* try opening it */
        fp = osfoprb(filepath, OSFTBIN);
    }

    /* return the result */
    return fp;
}

