#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/tads3/VMSTACK.CPP,v 1.3 1999/07/11 00:46:58 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmstack.cpp - VM stack implementation
Function
  
Notes
  
Modified
  10/28/98 MJRoberts  - Creation
*/

#include "t3std.h"
#include "vmtype.h"
#include "vmstack.h"
#include "vmfile.h"


/*
 *   allocate the stack 
 */
CVmStack::CVmStack(size_t max_depth)
{
    /* allocate the array of stack elements */
    arr_ = (vm_val_t *)t3malloc(max_depth * sizeof(arr_[0]));
    
    /* remember the maximum depth */
    max_depth_ = max_depth;

    /* nothing is currently on the stack */
    sp_ = arr_;
}

/*
 *   delete the stack 
 */
CVmStack::~CVmStack()
{
    /* delete the stack element array */
    t3free(arr_);
}

/* ------------------------------------------------------------------------ */
/*
 *   Write the stack to a saved state file 
 */
void CVmStack::write_to_file(VMG_ CVmFile *fp)
{
    vm_val_t *p;
    
    /* write the number of stack location's we're saving */
    fp->write_int2(sp_ - arr_);

    /* write each stack location as a DATAHOLDER value */
    for (p = arr_ ; p != sp_ ; ++p)
    {
        char buf[VMB_DATAHOLDER];

        /* store the portable DATAHOLDER value */
        vmb_put_dh_stk(vmg_ buf, p);

        /* write the buffer */
        fp->write_bytes(buf, VMB_DATAHOLDER);
    }
}

/*
 *   Read the stack from a saved state file 
 */
int CVmStack::read_from_file(VMG_ CVmFile *fp)
{
    size_t cnt;
    vm_val_t *p;
    
    /* read the number of stack locations to read */
    cnt = fp->read_uint2();

    /* make sure we have room */
    if (cnt > max_depth_)
        err_throw(VMERR_SAVED_STACK_OVERFLOW);

    /* read the elements */
    for (p = arr_ ; cnt != 0 ; ++p, --cnt)
    {
        char buf[VMB_DATAHOLDER];

        /* read this element's data */
        fp->read_bytes(buf, VMB_DATAHOLDER);

        /* translate and store the value */
        vmb_get_dh_stk(vmg_ buf, p);
    }

    /* set the stack pointer to the next open slot */
    sp_ = p;

    /* success */
    return 0;
}

